/* Program to solve Sudoku puzzles
 *
 * DL Ormand
 * Oct 2005
 */
#include <stdio.h>
#include <stdlib.h>

/* Puzzle contains three types of objects */
enum thing_type { ROW, COLUMN, BOX };

/* Each object instance in the puzzle is associated
 * with a row and a column in the table */
struct THING {
    enum thing_type type;
    int  row;
    int  column;
    int  index;
};

/* In a 9x9 puzzle, there are 9 rows, 9 cols, and 9 boxes */
#define NUMBER_OF_THINGS 27

/* Indicate where each object is in the table */
struct THING master_list[NUMBER_OF_THINGS] = {
    { ROW, 0, 0 },
    { ROW, 1, 0 },
    { ROW, 2, 0 },
    { ROW, 3, 0 },
    { ROW, 4, 0 },
    { ROW, 5, 0 },
    { ROW, 6, 0 },
    { ROW, 7, 0 },
    { ROW, 8, 0 },
    { COLUMN, 0, 0 },
    { COLUMN, 0, 1 },
    { COLUMN, 0, 2 },
    { COLUMN, 0, 3 },
    { COLUMN, 0, 4 },
    { COLUMN, 0, 5 },
    { COLUMN, 0, 6 },
    { COLUMN, 0, 7 },
    { COLUMN, 0, 8 },
    { BOX, 0, 0 },
    { BOX, 0, 3 },
    { BOX, 0, 6 },
    { BOX, 3, 0 },
    { BOX, 3, 3 },
    { BOX, 3, 6 },
    { BOX, 6, 0 },
    { BOX, 6, 3 },
    { BOX, 6, 6 }
};

/* Handy names for the objects in the puzzle, to locate their
 * description in the master list */
enum thing_id { ROW1=0, ROW2, ROW3, ROW4, ROW5, ROW6, ROW7, ROW8, ROW9,
                COL1, COL2, COL3, COL4, COL5, COL6, COL7, COL8, COL9,
                BOX1, BOX2, BOX3, BOX4, BOX5, BOX6, BOX7, BOX8, BOX9
};

/* Keep track of the order of plays */
struct PLAY {
    struct THING thing;
    int row;
    int column;
    int value;
};

#define MAX_PLAYS 81
struct PLAY play[MAX_PLAYS];
int play_index;

/* Playing field */
int table[9][9];
int start_table[9][9];

void reset(void);
void load_default_table(char *fn);
int  is_in_column(int i, int x, int y);
int  is_in_row(int i, int x, int y);
int  is_in_box(int i, int x, int y);
int  is_in_thing(int i, struct THING *t);
int  blanks_in_column(int x);
int  blanks_in_row(int y);
int  blanks_in_box(int x, int y);
int  blanks_in_table(void);
int  blanks_in_thing(struct THING *t);
int  next_blank_is_at(struct THING *t, int *px, int *py, int advance);
void print_table(void);

int main(int argv, char **argc)
{
    int x, y, i, min, flag, advance_flag;
    struct THING min_t;

    if(argv == 2) {
        load_default_table(argc[1]);
    }
    else {
        load_default_table(NULL);
    }
    reset();
    print_table();

    while(blanks_in_table()) {
        /* find the thing with the least number of blanks in it */
        min = 9;
        for(x=0; x<NUMBER_OF_THINGS; x++) {
            i=blanks_in_thing(&master_list[x]);
            if(i != 0 && i < min) {
                min = i;
                min_t = master_list[x];
                min_t.index = x;
            }
        }
#ifdef DEBUG
printf("Thing with the least number of blanks is index %d\n", min_t.index);
#endif

        /* Loop to find a number/space/thing set that works */
        i = 1;    /* Start at the first potential number for the thing */
        x = min_t.column;   /* Start at the root row/col for the thing */
        y = min_t.row;
        advance_flag = 0;
        do {
            /* get the next number required for the thing */
            flag = 1;
            for(; i<=9; i++) {
                if(!is_in_thing(i, &min_t)) {
                    flag = 0;
                    break;
                }
            }
#ifdef DEBUG
printf("Next number for this thing is %d\n", i);
#endif

            /* Find the first blank in the thing that this number
             * will fit in all other associated things */
            while(!flag && next_blank_is_at(&min_t, &x, &y, advance_flag)) {
#ifdef DEBUG
printf("Next blank for %d is at %d, %d\n", min_t.index, x, y);
#endif
                if(!is_in_row(i, x, y) &&
                   !is_in_column(i, x, y) &&
                   !is_in_box(i, x, y)) {
                    /* Put in table and last play */
                    table[x][y] = i;
                    play[play_index].thing = min_t;
                    play[play_index].row = y;
                    play[play_index].column = x;
                    play[play_index].value = i;
                    play_index++;
                    advance_flag = 0;
                    flag=1;
#ifdef DEBUG
printf("Put a %d at %d, %d, index=%d, play #%d\n",
        i, x, y, min_t.index, play_index-1);
#endif
                    break;
                }
                advance_flag = 1;
            }

            /* if unable to find a place, back out the last play */
            if(!flag) {
                play_index--;
                min_t = play[play_index].thing;
                y = play[play_index].row;
                x = play[play_index].column;
                i = play[play_index].value;
                table[x][y] = 0;
                advance_flag = 1;
#ifdef DEBUG
printf("Backing out %d at %d, %d, index=%d, play #%d\n",
        i, x, y, min_t.index, play_index+1);
#endif
            }
#ifdef DEBUG
printf("%d plays, index=%d\n", play_index, min_t.index);
#endif
        } while(!flag);
#ifdef DEBUG
        print_table();
/*        getchar(); */
#endif
    }

    printf("Done!!!\n\n");
    print_table();

    return 0;
}

void reset(void)
{
    int x, y;

    for(x=0; x<9; x++) {
        for(y=0; y<9; y++) {
            table[x][y] = start_table[x][y];
        }
    }

    play_index = 0;
}

void load_default_table(char *fn)
{
    int x, y, i;
    FILE *f;

    if(fn != NULL) {
        if((f = fopen(fn, "r")) == NULL) {
            fprintf(stderr, "puzzle: cannot open %s as input file\n", fn);
            exit(0);
        }
        for(y=0; y<9; y++) {
            for(x=0; x<9; x++) {
                do {
                    i = getc(f);
                    if(i == EOF) {
                        i = '0';
                    }
                    i = i - '0';
                } while(i < 0 || i > 10);
                start_table[x][y] = i;
            }
        }
    }

    else {
        for(x=0; x<9; x++) {
            for(y=0; y<9; y++) {
                start_table[x][y] = 0;
            }
        }

        start_table[0][1] = 5;
        start_table[0][3] = 6;
        start_table[0][5] = 9;
        start_table[0][6] = 3;
        start_table[0][8] = 4;
        start_table[1][3] = 1;
        start_table[1][4] = 7;
        start_table[1][5] = 2;
        start_table[1][8] = 5;
        start_table[2][7] = 2;
        start_table[3][1] = 7;
        start_table[3][2] = 4;
        start_table[3][5] = 1;
        start_table[4][0] = 3;
        start_table[4][8] = 1;
        start_table[5][3] = 5;
        start_table[5][6] = 6;
        start_table[5][7] = 4;
        start_table[6][1] = 2;
        start_table[7][0] = 7;
        start_table[7][3] = 2;
        start_table[7][4] = 4;
        start_table[7][5] = 5;
        start_table[8][0] = 1;
        start_table[8][2] = 8;
        start_table[8][3] = 3;
        start_table[8][5] = 7;
        start_table[8][7] = 6;
    }
}

void print_table(void)
{
    int x, y;

    for(y=0; y<9; y++) {
        if(y == 3 || y == 6) {
            printf("\n");
        }
        for(x=0; x<9; x++) {
            if(x == 3 || x == 6) {
                printf("| ");
            }
            if(table[x][y] == 0) {
                printf("|   ");
            }
            else {
                printf("| %d ", table[x][y]);
            }
        }
        printf("\n");
    }
}

/* Determine if the number i is in a column object that includes
 * row x and column y */
int  is_in_column(int i, int x, int y)
{
    for(y=0; y<9; y++) {
        if(table[x][y] == i)
            return 1;
    }
    return 0;
}

/* Determine if the number i is in a row object that includes
 * row x and column y */
int  is_in_row(int i, int x, int y)
{
    for(x=0; x<9; x++) {
        if(table[x][y] == i)
            return 1;
    }
    return 0;
}

/* Determine if the number i is in a box object that includes
 * row x and column y */
int  is_in_box(int i, int x, int y)
{
    int xx, yy;

    /* Find upper left corner of box that contains this x, y */
    x = (x / 3) * 3;
    y = (y / 3) * 3;
    
    for(xx=x; xx<(x+3); xx++) {
        for(yy=y; yy<(y+3); yy++) {
            if(table[xx][yy] == i)
                return 1;
        }
    }
    return 0;
}

/* Determine if the number i is in the given object */
int  is_in_thing(int i, struct THING *t)
{
    int retval;

    switch(t->type) {
        case ROW :
            retval = is_in_row(i, t->column, t->row);
            break;
        case COLUMN :
            retval = is_in_column(i, t->column, t->row);
            break;
        case BOX :
            retval = is_in_box(i, t->column, t->row);
            break;
    }

    return retval;
}

/* Count the number of blanks in a row object */
int  blanks_in_row(int y)
{
    int i, x;

    for(i=0, x=0; x<9; x++) {
        if(table[x][y] == 0)
            i++;
    }
    return i;
}

/* Count the number of blanks in a col object */
int  blanks_in_column(int x)
{
    int i, y;

    for(i=0, y=0; y<9; y++) {
        if(table[x][y] == 0)
            i++;
    }
    return i;
}

/* Count the number of blanks in a box object */
int  blanks_in_box(int x, int y)
{
    int i, xx, yy;

    for(i=0, xx=x; xx<(x+3); xx++) {
        for(yy=y; yy<(y+3); yy++) {
            if(table[xx][yy] == 0)
                i++;
        }
    }
    return i;
}

/* Count the number of blanks in the entire table
 * (Puzzle is done when there are none) */
int  blanks_in_table(void)
{
    int i, x;

    for(i=0, x=0; x<9; x++) {
        i += blanks_in_column(x);
    }

    return i;
}

/* Count the number of blanks in the given object */
int  blanks_in_thing(struct THING *t)
{
    int retval;

    switch(t->type) {
        case ROW :
            retval = blanks_in_row(t->row);
            break;
        case COLUMN :
            retval = blanks_in_column(t->column);
            break;
        case BOX :
            retval = blanks_in_box(t->column, t->row);
            break;
    }

    return retval;
}

/* Find row/col of next space in the given object */
int next_blank_is_at(struct THING *t, int *px, int *py, int advance_flag)
{
    int x, y, xx, yy, retval;

    x = *px;
    y = *py;

    switch(t->type) {
        case ROW :
            if(advance_flag) {
                x++;
            }
            for(retval=0; x<9; x++) {
                if(table[x][y] == 0) {
                    retval = 1;
                    break;
                }
            }
            break;
        case COLUMN :
            if(advance_flag) {
                y++;
            }
            for(retval=0; y<9; y++) {
                if(table[x][y] == 0) {
                   retval = 1;
                   break;
                }
            }
            break;
        case BOX :
            xx = ((x / 3) * 3) + 3;
            yy = ((y / 3) * 3) + 3;
            if(advance_flag) {
                y++;
                if(y == 3 || y == 6 || y == 9) {
                    x++;
                    y -= 3;
                }
            }
            for(retval=0; x<xx; x++) {
                for(; y<yy; y++) {
                    if(table[x][y] == 0) {
                        retval = 1;
                        break;
                    }
                }
                if(retval) {
                    break;
                }
                y = yy - 3;
            }
            break;
    }

    if(retval) {
        *px = x;
        *py = y;
    }
    return retval;
}

